<?php

declare(strict_types=1);

namespace Brick\VarExporter\Tests\Classes;

/**
 * A constructor, but no properties.
 */
class ConstructorAndNoProperties
{
    public function __construct()
    {
    }
}
